import collections as cs, functools as ft
import os, sys, io, enum, unittest, json, tempfile

import yaml

try: import pyaml.cli
except ImportError:
	sys.path.insert(1, os.path.join(__file__, *['..']*3))
	import pyaml.cli


class test_const(enum.IntEnum):
	dispatch = 2455
	heartbeat = 123

data = dict(
	key='value',
	path='/some/path',
	query_dump=dict(
		key1='тест1',
		key2='тест2',
		key3='тест3',
		последний=None ),
	ids=dict(),
	a=[1,None,'asd', 'не-ascii'], b=3.5, c=None )
data['query_dump_clone'] = data['query_dump']
data['ids']['id в уникоде'] = [4, 5, 6]
data['ids']['id2 в уникоде'] = data['ids']['id в уникоде']
data["'asd'\n!\0\1"] = dict(b=1, a=2)


class CliToolTests(unittest.TestCase):

	def data_hash(self, data):
		return json.dumps(data, sort_keys=True)

	def pyaml_dump_corrupted(self, dump, *args, append=None, **kws):
		out = dump(*args, **kws)
		if append: out += append
		return out

	def test_success(self):
		d, out, err = data.copy(), io.StringIO(), io.StringIO()

		ys = yaml.safe_dump(d)
		pyaml.cli.main( argv=list(),
			stdin=io.StringIO(ys), stdout=out, stderr=err )
		yaml.safe_load(out.getvalue())
		self.assertGreater(len(out.getvalue()), 150)
		self.assertEqual(err.getvalue(), '')

		d.update(
			d=test_const.heartbeat,
			asd=cs.OrderedDict(b=1, a=2) )
		ys = pyaml.dump(d)
		pyaml.cli.main( argv=list(),
			stdin=io.StringIO(ys), stdout=out, stderr=err )
		yaml.safe_load(out.getvalue())
		self.assertGreater(len(out.getvalue()), 150)
		self.assertEqual(err.getvalue(), '')

	def test_load_fail(self):
		d, out, err = data.copy(), io.StringIO(), io.StringIO()
		ys = yaml.safe_dump(d) + '\0asd : fgh : ghj\0'
		with self.assertRaises(yaml.YAMLError):
			pyaml.cli.main( argv=list(),
				stdin=io.StringIO(ys), stdout=out, stderr=err )

	def test_out_broken(self):
		d, out, err = data.copy(), io.StringIO(), io.StringIO()
		pyaml_dump, pyaml.dump = pyaml.dump, ft.partial(
			self.pyaml_dump_corrupted, pyaml.dump, append='\0asd : fgh : ghj\0' )
		try:
			ys = yaml.safe_dump(d)
			pyaml.cli.main( argv=list(),
				stdin=io.StringIO(ys), stdout=out, stderr=err )
			with self.assertRaises(yaml.YAMLError):
				yaml.safe_load(out.getvalue())
			self.assertGreater(len(out.getvalue()), 150)
			self.assertRegex(err.getvalue(), r'^WARNING:')
		finally: pyaml.dump = pyaml_dump

	def test_out_mismatch(self):
		d, out, err = data.copy(), io.StringIO(), io.StringIO()
		pyaml_dump, pyaml.dump = pyaml.dump, ft.partial(
			self.pyaml_dump_corrupted, pyaml.dump, append='\nextra-key: value' )
		try:
			ys = yaml.safe_dump(d)
			pyaml.cli.main( argv=list(),
				stdin=io.StringIO(ys), stdout=out, stderr=err )
			yaml.safe_load(out.getvalue())
			self.assertGreater(len(out.getvalue()), 150)
			self.assertRegex(err.getvalue(), r'^WARNING:')
		finally: pyaml.dump = pyaml_dump

	def test_out_err_nocheck(self):
		d, out, err = data.copy(), io.StringIO(), io.StringIO()
		pyaml_dump, pyaml.dump = pyaml.dump, ft.partial(
			self.pyaml_dump_corrupted, pyaml.dump, append='\0asd : fgh : ghj\0' )
		try:
			ys = yaml.safe_dump(d)
			pyaml.cli.main( argv=['-q'],
				stdin=io.StringIO(ys), stdout=out, stderr=err )
			with self.assertRaises(yaml.YAMLError):
				yaml.safe_load(out.getvalue())
			self.assertGreater(len(out.getvalue()), 150)
			self.assertEqual(err.getvalue(), '')
		finally: pyaml.dump = pyaml_dump

	def test_replace(self):
		d, out, err = data.copy(), io.StringIO(), io.StringIO()
		sys_out, sys_err = sys.stdout, sys.stderr
		with self.assertRaises(SystemExit):
			sys.stdout, sys.stderr = out, err
			try:
				pyaml.cli.main( argv=['-r'],
					stdin=io.StringIO(), stdout=out, stderr=err )
			finally: sys.stdout, sys.stderr = sys_out, sys_err
		self.assertEqual(out.getvalue(), '')
		self.assertGreater(len(err.getvalue()), 50)
		err.seek(0); err.truncate()

		with tempfile.NamedTemporaryFile(prefix='.pyaml.test.') as tmp:
			d_json, d_yaml = json.dumps(d).encode(), pyaml.dump(d, bytes)
			tmp.write(d_json); tmp.flush()
			os.fchmod(tmp.fileno(), 0o1510)
			stat_tmp = os.fstat(tmp.fileno())

			pyaml.cli.main( argv=['-r', tmp.name],
				stdin=io.StringIO(), stdout=out, stderr=err )
			with open(tmp.name, 'rb') as tmp_new:
				d_new, stat_new = tmp_new.read(), os.fstat(tmp_new.fileno())
			self.assertEqual(out.getvalue(), '')
			self.assertEqual(err.getvalue(), '')

			tmp.seek(0); d_tmp = tmp.read()
			self.assertEqual(d_tmp, d_json)
			self.assertNotEqual(d_tmp, d_new)
			self.assertNotIn(d_json, d_new)
			self.assertEqual(yaml.safe_load(d_new), d)
			self.assertEqual(
				(stat_tmp.st_mode, stat_tmp.st_uid, stat_tmp.st_gid),
				(stat_new.st_mode, stat_new.st_uid, stat_new.st_gid) )

			os.chmod(tmp.name, 0o600)
			with open(tmp.name, 'r+') as tmp_new:
				tmp_new.write('\0asd : fgh : ghj\0')
				tmp_new.seek(0); d_new = tmp_new.read()
			with self.assertRaises(yaml.YAMLError):
				pyaml.cli.main( argv=['-r', tmp.name],
					stdin=io.StringIO(), stdout=out, stderr=err )
			self.assertEqual(out.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			with open(tmp.name, 'r') as tmp_new:
				tmp_new.seek(0); d_new2 = tmp_new.read()
			self.assertEqual(d_new, d_new2)


if __name__ == '__main__': unittest.main()



# For compatibility - pyproject.toml should work instead
# ...but it does not atm, so ends up being parsed to setup() values here anyway

import setuptools, re, pathlib as pl

toml_lines = (pl.Path(__file__).parent / 'pyproject.toml').read_text().split('\n')

def toml_lines_rollup(lines, n=0):
	if not lines[n:]: return
	if re.match('\s+', lines[n]):
		return [lines[n].strip(), *(toml_lines_rollup(lines, n+1) or list())]
	if tail := toml_lines_rollup(lines, n+1): lines[n] += ' '.join(tail)
toml_lines_rollup(toml_lines)

def toml_str(key):
	for line, line_next in zip(toml_lines, toml_lines[1:]):
		if (m := re.fullmatch(r'(\w[\w-]+)\s*=\s*(.*?)\s*', line)) and m[1] == key:
			return re.findall(r'"([^"]+?)"', m[2])
	else: raise KeyError(key)

setup_kws = dict(
	((k, k) for k in 'name version license description classifiers'.split()),
	url='Homepage', install_requires='dependencies' )
setup_kws = dict((k1, toml_str(k2)[0]) for k1, k2 in setup_kws.items())

setup_kws['keywords'] = toml_str('keywords')
setup_kws['author'], setup_kws['author_email'] = toml_str('authors')
if fp := getattr(setuptools, 'find_packages', None): setup_kws['packages'] = fp()

setuptools.setup(**setup_kws)



import os, sys, re, stat, json, tempfile, contextlib
import yaml, pyaml


@contextlib.contextmanager
def safe_replacement(path, *open_args, mode=None, xattrs=None, **open_kws):
	'Context to atomically create/replace file-path in-place unless errors are raised'
	path, xattrs = str(path), None
	if mode is None:
		try: mode = stat.S_IMODE(os.lstat(path).st_mode)
		except FileNotFoundError: pass
	if xattrs is None and getattr(os, 'getxattr', None): # MacOS
		try: xattrs = dict((k, os.getxattr(path, k)) for k in os.listxattr(path))
		except FileNotFoundError: pass
	open_kws.update( delete=False,
		dir=os.path.dirname(path), prefix=os.path.basename(path)+'.' )
	if not open_args: open_kws.setdefault('mode', 'w')
	with tempfile.NamedTemporaryFile(*open_args, **open_kws) as tmp:
		try:
			if mode is not None: os.fchmod(tmp.fileno(), mode)
			if xattrs:
				for k, v in xattrs.items(): os.setxattr(path, k, v)
			yield tmp
			if not tmp.closed: tmp.flush()
			try: os.fdatasync(tmp)
			except AttributeError: pass # MacOS
			os.rename(tmp.name, path)
		finally:
			try: os.unlink(tmp.name)
			except FileNotFoundError: pass


def main(argv=None, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
	import argparse, textwrap
	dd = lambda text: re.sub( r' \t+', ' ',
		textwrap.dedent(text).strip('\n') + '\n' ).replace('\t', '  ')
	parser = argparse.ArgumentParser(
		formatter_class=argparse.RawTextHelpFormatter,
		description='Process and dump prettified YAML to stdout.')
	parser.add_argument('path', nargs='?', metavar='path',
		help='Path to YAML to read (default: use stdin).')
	parser.add_argument('-r', '--replace', action='store_true',
		help='Replace specified path with prettified version in-place.')
	parser.add_argument('-w', '--width', type=int, metavar='chars', help=dd('''
		Max line width hint to pass to pyyaml for the dump.
		Only used to format scalars and collections (e.g. lists).'''))
	parser.add_argument('-v', '--vspacing', metavar='N[/M][g]', help=dd('''
		Custom thresholds for when to add vertical spacing (empty lines),
			to visually separate items in overly long YAML lists/mappings.
		"long" means both >split-lines in line-length and has >split-count items.
		Value has N[/M][g] format, with default being something like 40/2.
			N = min number of same-indent lines in a section to split.
			M = min count of values in a list/mapping to split.
			"g" can be added to clump single-line values at the top of such lists/maps.
		Values examples: 20g, 5/1g, 60/4, g, 10.'''))
	parser.add_argument('-q', '--quiet', action='store_true',
		help='Disable sanity-check on the output and suppress stderr warnings.')
	opts = parser.parse_args(sys.argv[1:] if argv is None else argv)

	if opts.replace and not opts.path:
		parser.error('-r/--replace option can only be used with a file path, not stdin')

	src = open(opts.path) if opts.path else stdin
	try: data = yaml.safe_load(src)
	finally: src.close()

	pyaml_kwargs = dict()
	if opts.width: pyaml_kwargs['width'] = opts.width
	if vspacing := opts.vspacing:
		if vspacing.endswith('g'):
			pyaml_kwargs['sort_dicts'] = pyaml.PYAMLSort.oneline_group
			vspacing = vspacing.strip('g')
		if vspacing:
			vspacing, (lines, _, count) = dict(), vspacing.strip().strip('/').partition('/')
			if lines: vspacing['split_lines'] = int(lines.strip())
			if count: vspacing['split_count'] = int(count.strip())
			pyaml_kwargs['vspacing'] = vspacing

	ys = pyaml.dump(data, **pyaml_kwargs)

	if not opts.quiet:
		try:
			data_chk = yaml.safe_load(ys)
			try: data_hash = json.dumps(data, sort_keys=True)
			except: pass # too complex for checking with json
			else:
				if json.dumps(data_chk, sort_keys=True) != data_hash:
					raise AssertionError('Data from before/after pyaml does not match')
		except Exception as err:
			p_err = lambda *a,**kw: print(*a, **kw, file=stderr, flush=True)
			p_err( 'WARNING: Failed to parse produced YAML'
				' output back to data, it is likely too complicated for pyaml' )
			err = f'[{err.__class__.__name__}] {err}'
			p_err('  raised error: ' + ' // '.join(map(str.strip, err.split('\n'))))

	if opts.replace:
		with safe_replacement(opts.path) as tmp: tmp.write(ys)
	else: stdout.write(ys)
